from django.db.models import (
    SET_NULL,
    FileField,
    ForeignKey,
    PositiveSmallIntegerField,
)

from edu_rdm_integration.enums import (
    CommandType,
)
from edu_rdm_integration.models import (
    AbstractCollectDataCommandProgress,
    AbstractExportDataCommandProgress,
)
from edu_rdm_integration.utils import (
    get_data_command_progress_attachment_path,
)


class EduRdmCollectDataCommandProgress(AbstractCollectDataCommandProgress):
    """
    Модель, хранящая данные для формирования и отслеживания асинхронных задач по сбору данных.
    """

    task = ForeignKey(
        to='async_task.RunningTask',
        verbose_name='Асинхронная задача',
        blank=True,
        null=True,
        on_delete=SET_NULL,
    )
    logs_link = FileField(
        upload_to=get_data_command_progress_attachment_path,
        max_length=255,
        verbose_name='Ссылка на файл логов',
    )
    type = PositiveSmallIntegerField(  # noqa: A003
        verbose_name='Тип команды',
        choices=CommandType.get_choices(),
    )

    class Meta(AbstractCollectDataCommandProgress.Meta):
        db_table = 'edu_rdm_collecting_data_command_progress'


class EduRdmExportDataCommandProgress(AbstractExportDataCommandProgress):
    """Команда экспорта данных."""

    task = ForeignKey(
        to='async_task.RunningTask',
        verbose_name='Асинхронная задача',
        null=True,
        blank=True,
        on_delete=SET_NULL,
    )
    logs_link = FileField(
        upload_to=get_data_command_progress_attachment_path,
        max_length=255,
        verbose_name='Файл лога',
    )
    type = PositiveSmallIntegerField(  # noqa: A003
        verbose_name='Тип команды',
        choices=CommandType.get_choices(),
    )

    class Meta(AbstractExportDataCommandProgress.Meta):
        db_table = 'edu_rdm_exporting_data_command_progress'



DELIMITER = ';'

# Разделитель между id подэтапа и сущностью в составном ключе очереди
REDIS_QUEUE_KEY_DELIMITER = ':'
# Ключ параметра Общий объем отправленных в РВД файлов
TOTAL_ATTACHMENTS_SIZE_KEY = 'rdm:total_attachments_size'

# Ограничение количества файлов на обработку
ATTACHMENTS_LIMIT = 500



# Generated by Django 3.1.14 on 2024-08-19 14:24

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0010_transferredentity_export_enabled'),
    ]

    operations = [
        migrations.CreateModel(
            name='ExportingDataSubStageEntity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('entity', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.regionaldatamartentityenum', verbose_name='Сущность РВД')),
                ('exporting_data_sub_stage', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='edu_rdm_integration.exportingdatasubstage', verbose_name='Подэтап выгрузки данных')),
            ],
            options={
                'verbose_name': 'Связь сущности и подэтапа выгрузки',
                'verbose_name_plural': 'Связи сущности и подэтапа выгрузки',
                'db_table': 'rdm_exporting_data_sub_stage_entity',
            },
        ),
    ]



# Generated by Django 2.2.28 on 2023-12-04 12:24
# flake8: noqa
import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('edu_rdm_integration', '0004_uploaderclientlog'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataMartRequestStatus',
            fields=[
                ('title', models.TextField(verbose_name='расшифровка значения')),
                ('value', models.IntegerField(verbose_name='значение')),
                ('key', models.CharField(db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ')),
                ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
            ],
            options={
                'verbose_name': 'Статус загрузки данных в Витрину',
                'verbose_name_plural': 'Статусы загрузки данных в Витрину',
                'db_table': 'rdm_request_status',
            },
        ),
        migrations.AlterField(
            model_name='exportingdatasubstageuploaderclientlog',
            name='request_id',
            field=models.CharField(blank=True, db_index=True, max_length=100, verbose_name='Id запроса загрузки в витрину'),
        ),
        migrations.AddField(
            model_name='uploadstatusrequestlog',
            name='request_status',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='edu_rdm_integration.DataMartRequestStatus', verbose_name='Статус запроса в витрине'),
        ),
    ]



import json
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Union,
)

from django.conf import (
    settings,
)
from redis import (
    Redis,
)

from edu_rdm_integration.export_data.consts import (
    REDIS_QUEUE_KEY_DELIMITER,
)
from edu_rdm_integration.export_data.dataclasses import (
    UploadFile,
)
from edu_rdm_integration.redis_cache import (
    as_text,
    get_redis_version,
)


class Queue(ABC):
    """Интерфейс очереди."""
    queue_key: str = ''

    @property
    @abstractmethod
    def count(self) -> int:
        """Возвращает кол-во всех элементов в очереди."""

    def is_empty(self) -> bool:
        """Возвращает признак пустая ли очередь."""
        return self.count == 0

    @abstractmethod
    def enqueue(self, *args, **kwargs) -> None:
        """Поместить в очередь."""

    @abstractmethod
    def dequeue(self) -> Any:
        """Вернуть из очереди."""

    @abstractmethod
    def clear(self) -> None:
        """Очистить очередь."""

    @abstractmethod
    def delete_from_queue(self, *args, **kwargs) -> None:
        """Удалить из очереди конкретное значение."""


class RdmRedisSubStageAttachmentQueue(Queue):
    """Очередь файлов и подэтапов.

    Данные хранятся следующим образом:
    - Подэтапы с сущностями (строка вида "407-MARKS" (sub_stage_id-entity)) хранятся в упорядоченном множестве Redis
    (Sorted Set in Redis)
    - Информация по файлам стандартно по ключу - ключом выступает sub_stage_id
    """
    queue_key = 'rdm:export_sub_stage_ids_queue'
    prefix = 'rdm:'

    def __init__(self, *args, **kwargs):
        """Инициализация объекта очереди Queue."""
        super().__init__(*args, **kwargs)

        self.connection = Redis(
            host=settings.RDM_REDIS_HOST,
            port=settings.RDM_REDIS_PORT,
            db=settings.RDM_REDIS_DB,
            password=settings.RDM_REDIS_PASSWORD
        )

    def _make_key(self, key: Union[int, str]) -> str:
        """Формирование ключа."""
        return f'{self.prefix}{key}'

    @property
    def count(self) -> int:
        """Возвращает количество подэтапов в очереди."""
        return self.connection.zcard(self.queue_key)

    def enqueue(self, stage_id, entity_name: str, attachmets: list[UploadFile]) -> None:
        """Помещение в очередь.

        Подэтап попадает в упорядоченную очередь.
        """
        stage_info = f'{stage_id}{REDIS_QUEUE_KEY_DELIMITER}{entity_name}'
        pipe = self.connection.pipeline()
        # Упрядочиваем подэтапы
        pipe.zadd(self.queue_key, {stage_info: stage_id})
        pipe.set(self._make_key(stage_id), json.dumps(attachmets))
        pipe.execute()

    def dequeue_sub_stage_attachments(self, sub_stage_id: int) -> list[UploadFile]:
        """Возвращает файлы подэтапа из кеша."""
        result = []
        attachments = self.connection.get(self._make_key(sub_stage_id))
        attachments = json.loads(attachments) if attachments else ()
        for attachment in attachments:
            result.append(UploadFile(*attachment))

        return result

    def dequeue(self) -> dict[tuple[Any, Any], list[UploadFile]]:
        """Возвращает подэтапы из очереди - берется вся очередь без ограничений."""
        upload_files = {}
        exported_sub_stages = self.connection.zrange(self.queue_key, 0, -1)
        for sub_stage in exported_sub_stages:
            sub_stage_info = as_text(sub_stage)
            sub_stage_id, sub_stage_entity = sub_stage_info.split(REDIS_QUEUE_KEY_DELIMITER)
            upload_files[(sub_stage_id, sub_stage_entity)] = self.dequeue_sub_stage_attachments(sub_stage_id)

        return upload_files

    def delete_sub_stages_attachments(self, sub_stage_id: int) -> None:
        """Удаляет информацию о файлах из кеша."""
        self.connection.delete(self._make_key(sub_stage_id))

    def delete_sub_stages_from_queue(self, sub_stage_id: int, entity_name: str) -> None:
        """Удаляет подэтап из очереди."""
        self.connection.zrem(self.queue_key, f'{sub_stage_id}{REDIS_QUEUE_KEY_DELIMITER}{entity_name}')

    def delete_from_queue(self, sub_stage_id: int, entity_name: str) -> None:
        """Удаление элемента из очереди."""
        self.delete_sub_stages_attachments(sub_stage_id)
        self.delete_sub_stages_from_queue(sub_stage_id, entity_name)

    def clear(self) -> int:
        """Удаление из очереди всех подэтапов."""
        script = """
            local prefix = "{0}"
            local q = KEYS[1]
            local count = 0
            while true do
                local stage = redis.call("zpopmin", q)
                local stage_id = stage[2]
                
                if stage_id == nil then
                    break
                end
                redis.call("del", prefix..stage_id)
                count = count + 1
            end
            return count
        """.format(self.prefix).encode('utf-8')

        script = self.connection.register_script(script)

        return script(keys=[self.queue_key])

    @property
    def connection_info(self) -> str:
        """Информация об используемом соединении."""
        version = '.'.join(map(str, get_redis_version(self.connection)))
        kwargs = self.connection.connection_pool.connection_kwargs
        host = kwargs['host']
        port = kwargs['port']
        db = kwargs['db']

        return f'Redis {version} on {host}:{port}/{db}'



from edu_rdm_integration.export_data.export import (
    BaseExportEntitiesData,
    ExportLatestEntitiesData,
)
from edu_rdm_integration.management.general import (
    BaseExportEntityDataCommand,
)


class Command(BaseExportEntityDataCommand):
    """
    Команда для экспорта данных за период с последней сборки до указанной даты.
    """

    # flake8: noqa: A003
    help = 'Команда для запуска функции экспорта данных для интеграции с "Региональная витрина данных"'

    def _prepare_export_entities_data_class(self, *args, **kwargs) -> BaseExportEntitiesData:
        """Возвращает объект класса экспорта данных сущностей РВД."""
        return ExportLatestEntitiesData(
            entities=kwargs.get('entities'),
            period_started_at=kwargs.get('period_started_at'),
            period_ended_at=kwargs.get('period_ended_at'),
            task_id=kwargs.get('task_id'),
            update_modified=kwargs.get('update_modified'),
        )



from datetime import (
    date,
    datetime,
    time,
    timedelta,
)

from django.conf import (
    settings,
)
from django.db.models import (
    OuterRef,
    Subquery,
    Value,
)
from django.db.models.functions import (
    Coalesce,
)
from django.db.transaction import (
    atomic,
)

from educommon import (
    logger,
)

from edu_rdm_integration.models import (
    CollectingDataStageStatus,
    CollectingDataSubStageStatus,
    CollectingExportedDataStage,
    CollectingExportedDataSubStage,
)


@atomic
def set_failed_status_suspended_collecting_data_stages() -> dict[str, int]:
    """Установить статус 'Завершено с ошибками' для зависших этапов и подэтапов сбора.

    Сборка считается зависшей в случае если за определенное в параметре RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT время,
    отсутствуют изменения в связанных подэтапах. Параметр RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT определяется
    в настройках приложения.
    """
    changed_status_result = {
        'change_stage_count': 0,
        'change_sub_stage_count': 0,
    }

    current_datetime = datetime.now()
    suspended_time_at = current_datetime - timedelta(minutes=settings.RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT)

    suspended_stage_ids = set(
        CollectingExportedDataStage.objects.annotate(
            last_sub_stage_started_at=Coalesce(
                Subquery(
                    CollectingExportedDataSubStage.objects.filter(
                        stage_id=OuterRef('pk')
                    ).values('started_at').order_by('-started_at')[:1]
                ),
                Value(datetime.combine(date.min, time.min))
            )
        ).filter(
            last_sub_stage_started_at__lt=suspended_time_at,
            status__in=(
                CollectingDataStageStatus.CREATED.key,
                CollectingDataStageStatus.IN_PROGRESS.key,
            ),
        ).values_list('pk', flat=True)
    )

    if suspended_stage_ids:
        logger.info(f'find suspended CollectingExportedDataStage: {", ".join(map(str, suspended_stage_ids))}..')

        change_stage_count = CollectingExportedDataStage.objects.filter(
            pk__in=suspended_stage_ids,
        ).update(
            status=CollectingDataStageStatus.FAILED.key,
            ended_at=current_datetime,
        )

        change_sub_stage_count = CollectingExportedDataSubStage.objects.filter(
            stage_id__in=suspended_stage_ids,
        ).update(
            status=CollectingDataSubStageStatus.FAILED.key,
            ended_at=current_datetime,
        )

        changed_status_result.update({
            'change_stage_count': change_stage_count,
            'change_sub_stage_count': change_sub_stage_count,
        })

    return changed_status_result



from m3_ext.ui.all_components import (
    ExtButton,
)
from objectpack.ui import (
    BaseListWindow,
)

from educommon.utils.ui import (
    append_template_globals,
)


class TransferredEntityListWindow(BaseListWindow):
    """Окно реестра сущностей для сбора и экспорта данных."""

    def _init_components(self):
        """Инициализация компонентов окна."""
        super()._init_components()

        self.export_off_button = ExtButton(
            text='Отключить экспорт', handler='offExport'
        )
        self.export_on_button = ExtButton(
            text='Включить экспорт', handler='onExport'
        )

    def _do_layout(self):
        """Размещение компонентов окна на форме."""
        super()._do_layout()

        self.grid.top_bar.items.extend((self.export_off_button, self.export_on_button))

    def set_params(self, params, *args, **kwargs):
        """Настройка окна."""
        super().set_params(params, *args, **kwargs)

        append_template_globals(self, 'ui-js/transferred-entity-list.js')
        self.export_change_action_url = (
            params['pack'].export_change_action.get_absolute_url()
        )
        self.pack = params['pack']



from abc import (
    ABCMeta,
)

from edu_rdm_integration.collect_data.base.functions import (
    BaseCollectingCalculatedDataFunction,
)
from edu_rdm_integration.collect_data.calculated.base.helpers import (
    BaseCollectingCalculatedExportedDataFunctionHelper,
)
from edu_rdm_integration.collect_data.calculated.base.results import (
    BaseCollectingCalculatedExportedDataFunctionResult,
)
from edu_rdm_integration.collect_data.calculated.base.validators import (
    BaseCollectingCalculatedExportedDataFunctionValidator,
)


class BaseCollectingCalculatedExportedDataFunction(BaseCollectingCalculatedDataFunction, metaclass=ABCMeta):
    """
    Базовый класс функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_helper_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionHelper]:
        """
        Возвращает класс помощника функции.
        """
        return BaseCollectingCalculatedExportedDataFunctionHelper

    def _prepare_validator_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionValidator]:
        """
        Возвращает класс валидатора функции.
        """
        return BaseCollectingCalculatedExportedDataFunctionValidator

    def _prepare_result_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionResult]:
        """
        Возвращает класс результата функции.
        """
        return BaseCollectingCalculatedExportedDataFunctionResult



import logging
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Optional,
)


class BaseOperationData(ABC):
    """Базовый класс операций с данными."""

    def __init__(self, **kwargs):
        # Идентификатор команды для передачи сигналу manager_created
        self.command_id: Optional[int] = kwargs.get('command_id')

        self._file_handler: Optional[logging.FileHandler] = None

        self._add_file_handler()

    @property
    @abstractmethod
    def _log_file_path(self) -> str:
        """Путь до лог файла."""

    def _add_file_handler(self) -> None:
        """Добавляет обработчик логов."""
        if self.command_id:
            self._file_handler = logging.FileHandler(self._log_file_path)

            logging.getLogger('info_logger').addHandler(self._file_handler)
            logging.getLogger('exception_logger').addHandler(self._file_handler)

    def _remove_file_handler(self) -> None:
        """Удаляет обработчик логов."""
        if self._file_handler:
            logging.getLogger('info_logger').removeHandler(self._file_handler)
            logging.getLogger('exception_logger').removeHandler(self._file_handler)

            self._file_handler.close()



from edu_rdm_integration.collect_data.base.helpers import (
    BaseCollectingDataFunctionHelper,
    BaseCollectingDataRunnerHelper,
)
from edu_rdm_integration.collect_data.calculated.base.caches import (
    BaseCollectingCalculatedExportedDataFunctionCacheStorage,
    BaseCollectingCalculatedExportedDataRunnerCacheStorage,
)


class BaseCollectingCalculatedExportedDataRunnerHelper(BaseCollectingDataRunnerHelper):
    """
    Базовый класс помощников ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_cache_class(self) -> type[BaseCollectingCalculatedExportedDataRunnerCacheStorage]:
        """
        Возвращает класс кеша помощника ранера.
        """
        return BaseCollectingCalculatedExportedDataRunnerCacheStorage


class BaseCollectingCalculatedExportedDataFunctionHelper(BaseCollectingDataFunctionHelper):
    """
    Базовый класс помощников функций сбора расчетных данных для интеграции с "Региональная витрина данных".
    """

    def _prepare_cache_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionCacheStorage]:
        """
        Возвращает класс кеша помощника функции.
        """
        return BaseCollectingCalculatedExportedDataFunctionCacheStorage



from codecs import (
    open,
)
from pathlib import (
    Path,
)

import pip
from packaging.version import (
    Version,
)
from setuptools import (
    find_packages,
    setup,
)


PROJECT = 'edu_rdm_integration'

current_dir_path = Path().resolve()


#  Получение полного описания
with open(str(current_dir_path / 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(str(current_dir_path / 'CHANGELOG.md'), encoding='utf-8') as f:
    long_description += f.read()


production_requirements_path = current_dir_path / 'requirements' / 'production.txt'


if hasattr(pip, '__version__') and Version(str(pip.__version__)) >= Version('20'):
    from pip._internal.network.session import (
        PipSession,
    )
    from pip._internal.req import (
        parse_requirements,
    )

    requirements = parse_requirements(str(production_requirements_path), session=PipSession())
elif (
    hasattr(pip, '__version__')
    and Version('10.0.0') <= Version(str(pip.__version__)) <= Version('19.3.1')
):
    from pip._internal import (
        download,
        req,
    )

    requirements = req.parse_requirements(str(production_requirements_path), session=download.PipSession())
elif hasattr(pip, '__version__') and Version(str(pip.__version__)) >= Version('7.0'):
    requirements = pip.req.parse_requirements(str(production_requirements_path), session=pip.download.PipSession())
else:
    requirements = pip.req.parse_requirements(str(production_requirements_path))

install_requires = [str(item.requirement) for item in requirements]

setup(
    name=PROJECT.replace('_', '-'),

    description='Интеграция с Региональной витриной данных',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='BARS Group',
    author_email='bars@bars.group',

    url='',
    download_url='',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
    ],

    platforms=['Any'],

    scripts=[],

    provides=[],

    namespace_packages=[],
    package_dir={'': 'src'},
    packages=find_packages('src', exclude=('tests', 'tests.*')),
    include_package_data=True,

    package_data={
        '': [
            '*.conf',
            '*.tmpl',
            '*.sh',
            'Dockerfile',
            '*.yaml',
        ],
    },

    install_requires=install_requires,

    zip_safe=False,
)
